#include <stubmanagerframework.h>
#include <vector>
#include <queue>
#include <set>
#include <cassert>

using namespace std;

constexpr int SCHEDULE = 946149565,
                 VISIT = 547293220,
                ANSWER = 345685428;

#ifdef DEBUG
	#define TOO_MANY_QUERIES "Too many queries"
	#define INVALID_QUERY "Invalid call to schedule"
	#define WRONG_ANSWER "Wrong answer"
	#define INVALID_ANSWER "Invalid arguments to answer"
	#define NO_ANSWER "No answer"
	#define CORRECT "Correct"
	#define SECURITY_VIOLATION "Security violation!"
#else
	#define TOO_MANY_QUERIES "Not correct"
	#define INVALID_QUERY "Not correct"
	#define WRONG_ANSWER "Not correct"
	#define INVALID_ANSWER "Not correct"
	#define NO_ANSWER "Not correct"
	#define CORRECT "Correct"
	#define SECURITY_VIOLATION "Security violation!"
#endif

int N, V, T;
double P;
vector<vector<int>> A;

set<int> queryIndices;
vector<int> queryResult;
vector<pair<int, int>> sortingNetwork;

// Return true with a probability of approximately P
bool prob() {
	return rand() < P + (P * RAND_MAX);
}

int read_int_or_fail() {
	int i;
	if (fscanf(fcommin, "%d", &i) != 1)
		result(0.0, SECURITY_VIOLATION);
	return i;
}

void handle_schedule() {
	int i = read_int_or_fail() - 1;
	if (i < 0 || i >= N || queryIndices.find(i) != queryIndices.end())
		result(0.0, INVALID_QUERY);
	queryIndices.insert(i);

	int j = read_int_or_fail() - 1;
	if (j < 0 || j >= N || queryIndices.find(j) != queryIndices.end())
		result(0.0, INVALID_QUERY);
	queryIndices.insert(j);

	// Answer yes with probability P if P \in [0, 1]
	bool res = prob();
	// Otherwise, answer yes or no according to the first permutation
	if (P < -0.5)
		res = A[i][0] < A[j][0];
	queryResult.push_back(res);
	// Sort the permutation such that the answer is yes iff A[i] < A[j]
	if (!res)
		swap(i, j);
	sortingNetwork.emplace_back(i, j);
}

void handle_visit() {
	V--;
	if (V < 0)
		result(0.0, TOO_MANY_QUERIES);

	fprintf(fcommout, "%d", (int) queryResult.size());
	for (int i = 0; i < (int) queryResult.size(); i++)
		fprintf(fcommout, " %d", queryResult[i]);
	fprintf(fcommout, "\n");
	fflush(fcommout);

	queryIndices.clear();
	queryResult.clear();
}

void handle_answer() {
	while (!queryResult.empty()) {
		queryResult.pop_back();
		sortingNetwork.pop_back();
	}

	// Perform the sorting operations from the solution
	for (pair<int, int> p : sortingNetwork) {
		int i = p.first, j = p.second;
		for (int t = 0; t < T; t++)
			if (A[i][t] > A[j][t])
				swap(A[i][t], A[j][t]);
	}

	// Check that the answer matches with all permutations
	for (int i = 0; i < N; i++) {
		int a = read_int_or_fail() - 1;
		if (a < 0 || a >= N)
			result(0.0, INVALID_ANSWER);
		for (int t = 0; t < T; t++)
			if (A[a][t] != i + 1)
				result(0.0, WRONG_ANSWER);
	}

	// In case no swaps were performed, check that the answer is unique for the performed sorting operations
	if (P < 0) {
		vector<int> d(N), dp(N, 1);
		vector<vector<int>> e(N);
		for (pair<int, int> p : sortingNetwork) {
			d[p.second]++;
			e[p.first].push_back(p.second);
		}
		queue<int> q;
		for (int i = 0; i < N; i++)
			if (d[i] == 0)
				q.push(i);
		int ldp = 0;
		while (!q.empty()) {
			int i = q.front();
			q.pop();
			ldp = dp[i];
			for (int j : e[i]) {
				dp[j] = max(dp[j], dp[i] + 1);
				d[j]--;
				if (d[j] == 0)
					q.push(j);
			}
		}
		if (ldp != N)
			result(0.0, WRONG_ANSWER);
	}

	result(1.0, CORRECT);
}

void check() {
	message_on_shutdown = true;

	int S, TE, TR;
	assert(fscanf(fin, "%d %d %d %d %d %lf", &N, &V, &S, &TE, &TR, &P) == 6);
	srand(S);
	T = TE + TR;

	A.resize(N);
	// Read TE explicit permutations from generator
	for (int t = 0; t < TE; t++)
		for (int i = 0; i < N; i++) {
			int a;
			assert(fscanf(fin, "%d", &a) == 1);
			A[i].push_back(a);
		}

	vector<int> a(N);
	for (int i = 0; i < N; i++)
		a[i] = i + 1;
	// Generate TR permutations randomly
	for (int t = 0; t < TR; t++) {
		random_shuffle(a.begin(), a.end());
		for (int i = 0; i < N; i++)
			A[i].push_back(a[i]);
	}

	fprintf(fcommout, "%d %d\n", N, V);
	fflush(fcommout);

	for (;;) {
		int type;
		int fscanf_return = fscanf(fcommin, "%d", &type);

		if (fscanf_return == EOF)
			break;
		else if (fscanf_return != 1)
			result(0.0, SECURITY_VIOLATION);

		switch(type) {
			case SCHEDULE:
				handle_schedule();
				break;

			case VISIT:
				handle_visit();
				break;

			case ANSWER:
				handle_answer();
				break;

			default:
				result(0.0, SECURITY_VIOLATION);
				break;
		}
	}

	result(0.0, NO_ANSWER);
}
